#include "CdcUpdater.h"
#include "BoostLog.h"
#include "StringUtility.h"
#include <QDebug>
#include <QSerialPort>
#include <filesystem>
#include <fstream>
#include <mbedtls/md5.h>
#include <sstream>

static uint8_t calculate_crc8(const uint8_t *data, int length) {
    static constexpr uint8_t crc8_table[256] = {
        0x00, 0x07, 0x0e, 0x09, 0x1c, 0x1b, 0x12, 0x15, 0x38, 0x3f, 0x36, 0x31, 0x24, 0x23, 0x2a, 0x2d, 0x70, 0x77,
        0x7e, 0x79, 0x6c, 0x6b, 0x62, 0x65, 0x48, 0x4f, 0x46, 0x41, 0x54, 0x53, 0x5a, 0x5d, 0xe0, 0xe7, 0xee, 0xe9,
        0xfc, 0xfb, 0xf2, 0xf5, 0xd8, 0xdf, 0xd6, 0xd1, 0xc4, 0xc3, 0xca, 0xcd, 0x90, 0x97, 0x9e, 0x99, 0x8c, 0x8b,
        0x82, 0x85, 0xa8, 0xaf, 0xa6, 0xa1, 0xb4, 0xb3, 0xba, 0xbd, 0xc7, 0xc0, 0xc9, 0xce, 0xdb, 0xdc, 0xd5, 0xd2,
        0xff, 0xf8, 0xf1, 0xf6, 0xe3, 0xe4, 0xed, 0xea, 0xb7, 0xb0, 0xb9, 0xbe, 0xab, 0xac, 0xa5, 0xa2, 0x8f, 0x88,
        0x81, 0x86, 0x93, 0x94, 0x9d, 0x9a, 0x27, 0x20, 0x29, 0x2e, 0x3b, 0x3c, 0x35, 0x32, 0x1f, 0x18, 0x11, 0x16,
        0x03, 0x04, 0x0d, 0x0a, 0x57, 0x50, 0x59, 0x5e, 0x4b, 0x4c, 0x45, 0x42, 0x6f, 0x68, 0x61, 0x66, 0x73, 0x74,
        0x7d, 0x7a, 0x89, 0x8e, 0x87, 0x80, 0x95, 0x92, 0x9b, 0x9c, 0xb1, 0xb6, 0xbf, 0xb8, 0xad, 0xaa, 0xa3, 0xa4,
        0xf9, 0xfe, 0xf7, 0xf0, 0xe5, 0xe2, 0xeb, 0xec, 0xc1, 0xc6, 0xcf, 0xc8, 0xdd, 0xda, 0xd3, 0xd4, 0x69, 0x6e,
        0x67, 0x60, 0x75, 0x72, 0x7b, 0x7c, 0x51, 0x56, 0x5f, 0x58, 0x4d, 0x4a, 0x43, 0x44, 0x19, 0x1e, 0x17, 0x10,
        0x05, 0x02, 0x0b, 0x0c, 0x21, 0x26, 0x2f, 0x28, 0x3d, 0x3a, 0x33, 0x34, 0x4e, 0x49, 0x40, 0x47, 0x52, 0x55,
        0x5c, 0x5b, 0x76, 0x71, 0x78, 0x7f, 0x6a, 0x6d, 0x64, 0x63, 0x3e, 0x39, 0x30, 0x37, 0x22, 0x25, 0x2c, 0x2b,
        0x06, 0x01, 0x08, 0x0f, 0x1a, 0x1d, 0x14, 0x13, 0xae, 0xa9, 0xa0, 0xa7, 0xb2, 0xb5, 0xbc, 0xbb, 0x96, 0x91,
        0x98, 0x9f, 0x8a, 0x8d, 0x84, 0x83, 0xde, 0xd9, 0xd0, 0xd7, 0xc2, 0xc5, 0xcc, 0xcb, 0xe6, 0xe1, 0xe8, 0xef,
        0xfa, 0xfd, 0xf4, 0xf3,
    };

    uint8_t crc = 0;
    for (int i = 0; i < length; ++i) {
        crc = crc8_table[crc ^ data[i]];
    }
    return crc;
}

CdcUpdater::CdcUpdater(QObject *parent) : QObject(parent) {
}

CdcUpdater::~CdcUpdater() {
}

void CdcUpdater::start(const QString &path, const QSerialPortInfo &info) {
    if (info.isNull()) {
        startSearchDevice();
    } else {
        open(info);
    }
#ifdef Q_OS_WINDOWS
    m_path = Amass::StringUtility::UTF8ToGBK(path.toStdString());
#else
    m_path = path.toStdString();
#endif
    LOG(info) << "ota file: " << m_path;
    emit progressChanged(++m_progress);
}

void CdcUpdater::startSearchDevice() {
    if (m_timerId < 0) {
        m_timerId = startTimer(1000);
    }
}

bool CdcUpdater::open(const QSerialPortInfo &info) {
    m_serialPort = std::make_shared<QSerialPort>(info);
    connect(m_serialPort.get(), &QSerialPort::readyRead, this, &CdcUpdater::onReadyRead);
    m_serialPort->setBaudRate(2000000);
    m_serialPort->setDataBits(QSerialPort::Data8);
    m_serialPort->setStopBits(QSerialPort::OneStop);

    auto ret = m_serialPort->open(QSerialPort::ReadWrite);
    if (ret) {
        ret = write(EnterUpgrade);
        emit progressChanged(++m_progress);
    }

    return ret;
}

bool CdcUpdater::write(Command command, const uint8_t *data, uint32_t size) {
    std::array<uint8_t, 8> packet = {0};
    packet[0] = 0xAA;
    packet[1] = 0x55;
    packet[2] = command;
    if (data != nullptr) {
        memcpy(&packet[3], data, size);
    }
    if (command != TransferBin) {
        std::ostringstream oss;
        for (int i = 0; i < packet.size(); i++) {
            oss << "0x" << std::hex << (static_cast<int>(packet[i]) & 0xff) << " ";
        }
        LOG(info) << "write " << oss.str();
    }
    return m_serialPort->write(reinterpret_cast<const char *>(packet.data()), packet.size());
}

std::optional<QSerialPortInfo> CdcUpdater::searchDevice() {
    std::optional<QSerialPortInfo> ret = std::nullopt;
    const auto serialPortInfos = QSerialPortInfo::availablePorts();
    for (const QSerialPortInfo &portInfo : serialPortInfos) {
        if ((portInfo.vendorIdentifier() == 0) && (portInfo.productIdentifier() == 0)) continue;
        LOG(info) << "portName:" << portInfo.portName().toStdString() << ", vendorIdentifier: 0x" << std::hex
                  << portInfo.vendorIdentifier() << ", productIdentifier: 0x" << std::hex
                  << portInfo.productIdentifier();
        if (portInfo.vendorIdentifier() == 0xffff) {
            ret = portInfo;
            break;
        }
    }
    return ret;
}

void CdcUpdater::timerEvent(QTimerEvent *event) {
    auto device = searchDevice();
    if (device) {
        LOG(info) << "founded device: " << device->portName().toStdString();
        emit deviceDiscovered(*device);
        killTimer(m_timerId);
        m_timerId = -1;
    }
    LOG(info) << "----------";
}

void CdcUpdater::transferBin() {
    char buffer[4096] = {0};
    m_ifs->read(buffer, sizeof(buffer));
    auto readSize = m_ifs->gcount();

    if (readSize > 0) {
        m_progress = m_progressBeforeTranster +
                     (99 - m_progressBeforeTranster) * static_cast<float>(m_packetIndex) / m_totalPackageSize;
        if (m_progress > 99) m_progress = 99;
        emit progressChanged(m_progress);

        std::array<uint8_t, 5> content;
        content[0] = (m_packetIndex >> 8) & 0xFF;
        content[1] = m_packetIndex & 0xFF;

        content[2] = (readSize >> 8) & 0xFF;
        content[3] = readSize & 0xFF;

        content[4] = calculate_crc8(reinterpret_cast<const uint8_t *>(buffer), readSize);
        write(TransferBin, content.data(), content.size());
        m_serialPort->write(buffer, readSize);

        m_packetIndex++;
    }
}

void CdcUpdater::onReadyRead() {
    auto data = m_serialPort->readAll();
    uint8_t command = static_cast<uint8_t>(data[2]);
    if (command != TransferBinReply) {
        std::ostringstream oss;
        for (int i = 0; i < data.size(); i++) {
            oss << "0x" << std::hex << (static_cast<int>(data[i]) & 0xff) << " ";
        }
        LOG(info) << "onReadyRead, size: " << data.size() << ", " << oss.str();
    }
    if (command == EnterUpgradeReply) {
        write(GetVersion);
        emit progressChanged(++m_progress);
        emit message("获取模组OTA版本......");
    } else if (command == GetVersionReply) {
        LOG(info) << "device ota version: 0x" << std::hex << (static_cast<int>(data[4]) & 0xff);
        int fileSize = std::filesystem::file_size(m_path);
        m_totalPackageSize = fileSize / 4096;
        if ((fileSize % 4096) > 0) m_totalPackageSize++;
        std::array<uint8_t, 3> content;
        content[0] = (m_totalPackageSize >> 8) & 0xFF;
        content[1] = m_totalPackageSize & 0xFF;
        content[2] = 0x14;
        write(StartUpgrade, content.data(), content.size());
        emit progressChanged(++m_progress);
        emit message("模组进入升级状态......");
    } else if (command == StartUpgradeReply) {
        std::ifstream ifs(m_path, std::ifstream::binary);
        char buffer[4096] = {0};
        mbedtls_md5_context context;
        mbedtls_md5_init(&context);
        mbedtls_md5_starts(&context);
        uint8_t md5[16];
        while (ifs) {
            ifs.read(buffer, sizeof(buffer));
            auto readSize = ifs.gcount();
            mbedtls_md5_update(&context, reinterpret_cast<uint8_t *>(buffer), readSize);
        }
        mbedtls_md5_finish(&context, md5);
        mbedtls_md5_free(&context);
        std::ostringstream oss;
        for (int i = 0; i < sizeof(md5); i++) {
            oss << "0x" << std::hex << (static_cast<int>(md5[i]) & 0xff) << " ";
        }
        LOG(info) << "md5: " << oss.str();
        write(Md5Checksum);
        m_serialPort->write(reinterpret_cast<const char *>(md5), sizeof(md5));
        emit progressChanged(++m_progress);
        emit message("发送升级包MD5校验值......");
    } else if (command == Md5ChecksumReply) {
        m_ifs = std::make_shared<std::ifstream>(m_path, std::ifstream::binary);
        m_packetIndex = 0;
        m_progressBeforeTranster = m_progress;
        transferBin();
        emit message("发送升级包数据......");
    } else if (command == TransferBinReply) {
        transferBin();
        if (m_packetIndex == m_totalPackageSize) {
            message("正在擦写flash,请稍后......");
        }
    } else if (command == CurrentVersion) {
        LOG(info) << "update finieshed.";
        write(CurrentVersionReply);
        // std::this_thread::sleep_for(std::chrono::seconds(1));
        write(ExitUpgrade);
        m_progress = 100;
        emit progressChanged(m_progress);
        emit updateFinished();
        emit message("升级结束。");
    } else if (command == UpgradeExit) {
        LOG(info) << "module exit boot mode.";
    }
}
